import path from 'path';
import { toArray } from '@unocss/core';
import rgba from 'color-rgba';
const remUnitRE = /(-?[\d.]+)rem(\s+!important)?;/;
const matchCssVarNameRE = /var\((?<cssVarName>--[^,|)]+)(?:,(?<fallback>[^)]+))?\)/g;
const cssColorRE = /(?:#|0x)(?:[a-f0-9]{3}|[a-f0-9]{6})\b|(?:rgb|hsl)a?\(.*\)/g;
const varFnRE = /var\((--[^,|)]+)(?:,([^)]+))?\)/;
export function throttle(func, timeFrame) {
    let lastTime = 0;
    let timer;
    return function (...args) {
        const now = Date.now();
        clearTimeout(timer);
        if (now - lastTime >= timeFrame) {
            lastTime = now;
            return func(...args);
        }
        else {
            timer = setTimeout(func, timeFrame, ...args);
        }
    };
}
export async function getCSS(uno, utilName) {
    const { css } = await uno.generate(new Set(toArray(utilName)), { preflights: false, safelist: false });
    return css;
}
/**
 * Credit to [@voorjaar](https://github.com/voorjaar)
 *
 * @see https://github.com/windicss/windicss-intellisense/issues/13
 * @param str
 */
export function addRemToPxComment(str, remToPixel = 16) {
    if (!str)
        return '';
    if (remToPixel < 1)
        return str;
    let index = 0;
    const output = [];
    while (index < str.length) {
        const rem = str.slice(index).match(remUnitRE);
        if (!rem || !rem.index)
            break;
        const px = ` /* ${Number.parseFloat(rem[1]) * remToPixel}px */`;
        const end = index + rem.index + rem[0].length;
        output.push(str.slice(index, end));
        output.push(px);
        index = end;
    }
    output.push(str.slice(index));
    return output.join('');
}
export async function getPrettiedCSS(uno, util, remToPxRatio) {
    const result = (await uno.generate(new Set(toArray(util)), { preflights: false, safelist: false }));
    const css = addRemToPxComment(result.css, remToPxRatio);
    const prettified = css;
    return {
        ...result,
        prettified,
    };
}
export async function getPrettiedMarkdown(uno, util, remToPxRatio) {
    return `\`\`\`css\n${(await getPrettiedCSS(uno, util, remToPxRatio)).prettified}\n\`\`\``;
}
function getCssVariables(code) {
    var _a, _b, _c;
    const regex = /(?<key>--[^\s:]+):(?<value>.+?)[!;]/g;
    const cssVariables = new Map();
    for (const match of code.matchAll(regex)) {
        const key = (_a = match.groups) === null || _a === void 0 ? void 0 : _a.key;
        if (key)
            cssVariables.set(key, (_c = (_b = match.groups) === null || _b === void 0 ? void 0 : _b.value) !== null && _c !== void 0 ? _c : '');
    }
    return cssVariables;
}
/**
 * Get CSS color string from CSS string
 *
 * @example Input with CSS var
 * ```css
 *.dark [border="dark\:gray-700"] {
 *  --un-border-opacity: 1;
 *  border-color: rgb(55 65 81 / var(--un-border-opacity));
 *}
 * ```
 * return `rgb(55 65 81 / 1)`
 *
 * @example Input with no-value CSS var and its fallback value
 * ```css
 *.bg-brand-primary {
 *  background-color: hsl(217 78% 51% / var(--no-value, 0.5));
 *}
 * ```
 * return `hsl(217 78% 51% / 0.5)`
 *
 * @example Input with no-value CSS var
 * ```css
 *.bg-brand-primary {
 *  background-color: hsl(217 78% 51% / var(--no-value));
 *}
 * ```
 * return `rgb(217 78% 51%)`
 *
 * @param str - CSS string
 * @returns The **first** CSS color string (hex, rgb[a], hsl[a]) or `undefined`
 */
export function getColorString(str) {
    var _a, _b, _c;
    console.log('getColorString str:', str);
    let colorString = (_a = str.match(cssColorRE)) === null || _a === void 0 ? void 0 : _a[0]; // e.g rgb(248 113 113 / var(--maybe-css-var))
    console.log('colorString:', colorString);
    if (!colorString)
        return;
    const cssVars = getCssVariables(str);
    // replace `var(...)` with its value
    for (const match of colorString.matchAll(matchCssVarNameRE)) {
        const matchedString = match[0];
        const cssVarName = (_b = match.groups) === null || _b === void 0 ? void 0 : _b.cssVarName;
        const fallback = (_c = match.groups) === null || _c === void 0 ? void 0 : _c.fallback;
        if (cssVarName && cssVars.get(cssVarName)) {
            // rgb(248 113 113 / var(--un-text-opacity)) => rgb(248 113 113 / 1)
            colorString = colorString.replaceAll(matchedString, () => {
                var _a, _b;
                let v = (_a = cssVars.get(cssVarName)) !== null && _a !== void 0 ? _a : matchedString;
                // resolve nested css var
                while (v && v.startsWith('var(')) {
                    const varName = (_b = v.match(varFnRE)) === null || _b === void 0 ? void 0 : _b[1];
                    if (!varName) {
                        v = '';
                        break;
                    }
                    v = cssVars.get(varName) || '';
                }
                return v || '1';
            });
        }
        else if (fallback) {
            // rgb(248 113 113 / var(--no-value, 0.5)) => rgb(248 113 113 / 0.5)
            colorString = colorString.replaceAll(matchedString, fallback);
        }
        // rgb(248 113 113 / var(--no-value)) => rgba(248 113 113)
        colorString = colorString.replaceAll(/,?\s+var\(--.*?\)/g, '');
    }
    // if (!(new TinyColor(colorString).isValid))
    //   return
    if (/\/\)/.test(colorString))
        colorString = colorString.replace(/ \/\)/g, '/ 1)');
    return convertToRGBA(colorString);
}
export function isSubdir(parent, child) {
    const relative = path.relative(parent, child);
    return relative && !relative.startsWith('..') && !path.isAbsolute(relative);
}
export function isFulfilled(result) {
    return result.status === 'fulfilled';
}
export function isRejected(result) {
    return result.status === 'rejected';
}
export function convertToRGBA(rgbColor) {
    console.log('rgbColor:', rgbColor);
    const match = rgba(rgbColor);
    if ((match === null || match === void 0 ? void 0 : match.length) === 4) {
        const [r, g, b, alpha] = match;
        return {
            red: r / 255,
            green: g / 255,
            blue: b / 255,
            alpha
        };
    }
    return;
}
